/*
  Copyright (c) 2010 Alistair Crooks (agc@pkgsrc.org)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
*/

#include <sys/types.h>

#include <inttypes.h>
#include <regex.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "mj.h"
#include "defs.h"

#define MJ_START		(0)

#define MJ_OPEN_BRACKET		(MJ_OBJECT + 1)
#define MJ_CLOSE_BRACKET	(MJ_OPEN_BRACKET + 1)
#define MJ_OPEN_BRACE		(MJ_CLOSE_BRACKET + 1)
#define MJ_CLOSE_BRACE		(MJ_OPEN_BRACE + 1)
#define MJ_COLON		(MJ_CLOSE_BRACE + 1)
#define MJ_COMMA		(MJ_COLON + 1)

/* return the token type, and start and finish locations in string */
static int
gettok(const char *s, int *from, int *to, int *tok)
{
	static regex_t	tokregex;
	regmatch_t	matches[15];
	static int	compiled;

	if (!compiled) {
	/* XXX - bound integers */
		compiled = 1;
		(void) regcomp(&tokregex,
			"[ \t\r\n]*(([+-]?[0-9]+(\\.[0-9]*)?([eE][-+][0-9]+)?)|"
			"(\"([^\"\\]|\\.)*\")|(null)|(false)|(true)|([][{}:,]))",
			REG_EXTENDED);
	}
	*from = *to;
	if (regexec(&tokregex, &s[*from], 15, matches, 0) != 0) {
		return *tok = -1;
	}
	*to = *from + (int)(matches[1].rm_eo);
	*tok = (matches[2].rm_so >= 0) ? MJ_NUMBER :
		(matches[5].rm_so >= 0) ? MJ_STRING :
		(matches[7].rm_so >= 0) ? MJ_NULL :
		(matches[8].rm_so >= 0) ? MJ_FALSE :
		(matches[9].rm_so >= 0) ? MJ_TRUE :
		(matches[10].rm_so < 0) ? -1 :
			(s[*from + matches[10].rm_so] == '[') ? MJ_OPEN_BRACKET :
			(s[*from + matches[10].rm_so] == ']') ? MJ_CLOSE_BRACKET :
			(s[*from + matches[10].rm_so] == '{') ? MJ_OPEN_BRACE :
			(s[*from + matches[10].rm_so] == '}') ? MJ_CLOSE_BRACE :
			(s[*from + matches[10].rm_so] == ':') ? MJ_COLON :
				MJ_COMMA;
	*from += (int)(matches[1].rm_so);
	return *tok;
}

static int
parse(const char *s, int *from, int *to, int *tok)
{
	int	i;

	while (gettok(s, from, to, tok) >= 0) {
		printf("from:%d, to:%d: type:%d, '%.*s'\n", *from, *to, *tok, (int)(*to - *from), &s[*from]);
		switch(*tok) {
		case MJ_OPEN_BRACKET:
			while (parse(s, from, to, tok) >= 0 && *tok != MJ_CLOSE_BRACKET) {
				if (*tok != MJ_COMMA) {
					(void) fprintf(stderr, "expected comma at '%s'\n", &s[*from]);
					break;
				}
				if (gettok(s, from, to, tok) < 0) {
					return *tok = -1;
				}
			}
			return gettok(s, from, to, tok);
		case MJ_OPEN_BRACE:
			for (i = 0 ; parse(s, from, to, tok) >= 0 && *tok != MJ_CLOSE_BRACE ; i++) {
				if (((i % 2) == 0 && *tok != MJ_COLON) || (i % 2) == 1 && *tok != MJ_COMMA) {
					break;
				}
			}
			return gettok(s, from, to, tok);
		default:
			gettok(s, from, to, tok);
		}
	}
	return *tok;
}

int
main(void)
{
	int	from;
	int	to;
	int	tok;

	from = to = 0;
	parse("{ \"name\": \"object1\", \"size\": 1786909, \"length\":1234, \"height\": 5678 }", &from, &to, &tok);
	from = to = 0;
	parse("[ \"name\", \"object1\", \"size\", 1786909, \"length\", 1234, \"height\", 5678 ]", &from, &to, &tok);
	from = to = 0;
	parse("{ \"name\": \"object1\", \"size\": 1786909, \"length\":1234, \"height\": 5678, \"width\": 12345678901234567890, \"transparent\":false }", &from, &to, &tok);
	from = to = 0;
	parse("{ \"header\":\"pub\", \"key bits\":2048, \"pka\":\"RSA (Encrypt or Sign)\", \"key id\":\"1b68dcfcc0596823\", \"birthtime\":1073907523, \"duration\":0, \"uids\":[ \"Alistair Crooks <alistair@hockley-crooks.com>\", \"Alistair Crooks <agc@pkgsrc.org>\", \"Alistair Crooks <agc@netbsd.org>\", \"Alistair Crooks <agc@alistaircrooks.com>\", \"Alistair Crooks (Yahoo!) <agcrooks@yahoo-inc.com>\", [ { \"pgp version\":4, \"header\":\"sub\", \"key size\":2048, \"pka\":\"RSA (Encrypt or Sign)\", \"signer\":\"1b68dcfcc0596823\", \"signtime\":1073907686} ] ] }", &from, &to, &tok);
	exit(0);
}
